#!/usr/bin/env python3

import fcntl
import logging
from time import time, sleep
from typing import Union

import serial

from delumowave.config.delumo import (DEVICE, COMMAND, EEPROM_SIZE)

_LOGGER = logging.getLogger(__name__)


def hex_repr(data: Union[int, bytes, list, tuple, set]) -> str:
    """
    Get hexadecimal representation of data (single int digit or
    bytes, list of int, tuple of int, set of int). This function is used for
    pretty logging.

    Parameters
    ----------
    data : Union[int, bytes, list, tuple, set]
        Data that should be represented in hex.

    Returns
    -------
    str
        Depending on the input data type function returns value could be
        represented in string of single hex int or string of list of int.

    Examples
    --------
    If input data has ``int`` type

    >>> hex_repr(18)
    '0x12'

    If input data has ``bytes`` type

    >>> hex_repr(b'\\x12\\x23\\x34')
    '[0x12, 0x23, 0x34]'

    If input data has ``list`` type

    >>> hex_repr([18, 35, 52])
    '[0x12, 0x23, 0x34]'

    """

    if isinstance(data, (bytes, list, tuple, set)) and \
            all(isinstance(x, int) for x in data):
        return f'[{", ".join([f"{b:#04x}" for b in data])}]'
    elif isinstance(data, int):
        return f'{data:#04x}'
    else:
        return ''


class DelumoWaveController:
    def __init__(self,
                 port_path: str = '/dev/ttyUSB0',
                 baudrate: int = 115200,
                 bytesize: int = serial.EIGHTBITS,
                 parity: int = serial.PARITY_NONE,
                 stopbits: int = serial.STOPBITS_ONE,
                 timeout: float = 3.0,
                 write_timeout: float = 3.0,
                 unlock_timeout: float = 5.0,
                 sleep_time: float = 0.1,
                 block: bool = False,
                 logging_level: Union[str, int, None] = None):
        """
        Base class for low-level communication with nodes. This class provide
        base methods to send command to nodes, read nodes' EEPROM and write to
        nodes' EEPROM via radio controller, connected by UART or USB.

        Parameters
        ----------
        port_path : str, optional
            Device name via which radio controller is connected: depending on
            operating system. e.g. ``/dev/ttyUSB0``, ``/dev/ttyAMA0``
            on GNU/Linux or ``COM3`` on Windows.
            By default, ``/dev/ttyAMA0`` is used.
        baudrate : int, optional
            Baud rate such as 9600 or 115200 etc. By default, 115200 is used.
        bytesize : int, optional
            Number of data bits. Possible values: ``FIVEBITS``, ``SIXBITS``,
            ``SEVENBITS``, ``EIGHTBITS``. By default, ``EIGHTBITS`` is used.
        parity : int, optional
            Enable parity checking. Possible values: ``PARITY_NONE``,
            ``PARITY_EVEN``, ``PARITY_ODD``, ``PARITY_MARK``, ``PARITY_SPACE``.
            By default, ``PARITY_NONE`` is used.
        stopbits : int, optional
            Number of stop bits. Possible values: ``STOPBITS_ONE``,
            ``STOPBITS_ONE_POINT_FIVE``, ``STOPBITS_TWO``.
            By default, ``STOPBITS_ONE`` is used.
        timeout : float, optional
            Set a read timeout value. By default, ``3.0`` is used.
        write_timeout : float, optional
            Set a write timeout value. By default, ``3.0`` is used.
        unlock_timeout : float, optional
            How long (in seconds) try to to open radio controller in case if it
            is blocked by other process. By default, 5 seconds.
        sleep_time : float, optional
            How often (in seconds) try to open radio controller in case if it is
            blocked by other process. By default, each 0.1 second.
        block : bool, optional
            Block radio controller and wait until a radio transmission and node
            processing are complete. By default, block is True.
        logging_level : Union[str, int, None], optional
            Available logging levels: ``DEBUG``, ``INFO``, ``WARNING``,
            ``ERROR``, ``CRITICAL``, ``NOTSET``. Also it could be written by
            numeric values.
            By default, logging is disabled.

        Raises
        ------
        ValueError
            Will be raised when parameter are out of range, e.g. baud rate,
            data bits.
        SerialException
            In case the device can not be found or can not be configured.

        See Also
        --------
        serial.serial_for_url : function to open serial port
        logging.setLevel : function to set logging level (see available
            constants).
        """

        if logging_level is None:
            _LOGGER.disabled = True
        elif isinstance(logging_level, str):
            if logging_level.upper() in ('DEBUG', 'INFO', 'WARNING', 'ERROR',
                                         'CRITICAL', 'NOTSET'):
                _LOGGER.setLevel(logging_level.upper())
        elif isinstance(logging_level, int):
            _LOGGER.setLevel(logging_level)
        else:
            _LOGGER.setLevel('ERROR')

        _LOGGER.debug(f'DelumoWaveController __init__() with Args: {locals()}')

        self._port_path = port_path
        self.__unlock_timeout = unlock_timeout
        self.__sleep_time = sleep_time
        self.__block = block
        self.__node_addr = None
        self.__procedure = None
        self.__node_version = None
        self.__request = None
        self.__request_length = None
        self.__expected_request_length = None
        self.__response = None
        self.__response_cmd = None
        self.__response_data = None
        self.__response_length = None
        self.__expected_response_length = None
        self.__eeprom_data_length = None
        self.__data_length = None
        self.__eeprom_addr = None
        self.__eeprom_data = None

        self._controller = serial.serial_for_url(url=port_path,
                                                 baudrate=baudrate,
                                                 parity=parity,
                                                 stopbits=stopbits,
                                                 bytesize=bytesize,
                                                 timeout=timeout,
                                                 write_timeout=write_timeout,
                                                 do_not_open=True)

        _LOGGER.info(f'Radio controller is ready: {self._controller}')

    @property
    def _node_addr(self) -> bytes:
        return self.__node_addr

    @_node_addr.setter
    def _node_addr(self, node_addr: Union[bytes, list, tuple]):
        if not isinstance(node_addr, (bytes, list, tuple)):
            raise TypeError(f'Node address must be {bytes}, {list} or {tuple}, '
                            f'not {type(node_addr)}')
        if not len(node_addr) == 3:
            raise ValueError(f'Node address must have size 3, '
                             f'not {len(node_addr)}')
        if not all((isinstance(x, int) and 0x00 <= x <= 0xFF) for x in node_addr):
            raise ValueError(f'{type(node_addr)} item must be {int} and be in '
                             f'range [0x00, 0xFF]')

        self.__node_addr = bytes(node_addr)

    @property
    def _procedure(self) -> int:
        return self.__procedure

    @_procedure.setter
    def _procedure(self, procedure: int):
        if not isinstance(procedure, int):
            raise TypeError(f'Procedure must be {int}, not {type(procedure)}')
        elif not 0x00 <= procedure <= 0xFF:
            raise ValueError(f'Procedure must be in range [0x00, 0xFF], '
                             f'not {procedure}')

        self.__procedure = procedure

    @property
    def _node_version(self) -> int:
        return self.__node_version

    @_node_version.setter
    def _node_version(self, node_version: int):
        if not isinstance(node_version, int):
            raise TypeError(f'Node version must be {int}, '
                            f'not {type(node_version)}')
        elif not 0x00 <= node_version <= 0xFF:
            raise ValueError(f'Node version must be in range [0x00, 0xFF], '
                             f'not {node_version}')

        self.__node_version = node_version

    @property
    def _request(self) -> bytes:
        return self.__request

    @_request.setter
    def _request(self, request: Union[bytes, list, tuple]):
        if not isinstance(request, (bytes, list, tuple)):
            raise TypeError(f'Node address must be {bytes}, {list} or {tuple}, '
                            f'not {type(request)}')
        if not all((isinstance(x, int) and 0x00 <= x <= 0xFF) for x in request):
            raise ValueError(f'{type(request)} item must be {int} and be in '
                             f'range [0x00, 0xFF]')

        self.__request = bytes(request)
        self.__expected_request_length = len(self._request)

        if request[4] == COMMAND['READ_EEPROM']:
            # If requested data length is 0x00, then it expected whole EEPROM
            # data from eeprom_addr to the end of EEPROM. Otherwise it expect
            # data_length bytes from addr_eemprom
            if self._data_length != 0x00:
                self.__eeprom_data_length = self._data_length
            else:
                self.__eeprom_data_length = EEPROM_SIZE - self._eeprom_addr

            #: Request length + requested data length + 1 byte of check sum
            self.__expected_response_length = self.__expected_request_length + self.__eeprom_data_length + 1

    @property
    def _request_length(self) -> int:
        return self.__request_length

    @_request_length.setter
    def _request_length(self, request_length: int):
        if request_length == self._expected_request_length:
            self.__request_length = request_length
        else:
            raise ValueError(f'Request length {request_length} is not equal '
                             f'expected request length '
                             f'{self._expected_request_length}')

    @property
    def _response(self) -> bytes:
        return self.__response

    @_response.setter
    def _response(self, response: Union[bytes, list, tuple]):
        if not isinstance(response, (bytes, list, tuple)):
            raise TypeError(f'Node address must be {bytes}, {list} or {tuple}, '
                            f'not {type(response)}')
        if not all((isinstance(x, int) and 0x00 <= x <= 0xFF) for x in response):
            raise ValueError(f'{type(response)} item must be {int} and be in '
                             f'range [0x00, 0xFF]')

        self.__response = bytes(response)
        self.__response_length = len(self._response)
        self.__response_cmd = self._response[:self._request_length]
        self.__response_data = self._response[self._request_length:]

        if self.__response_length != self._expected_response_length:
            # TODO: возможно ли другая причина исключения кроме времени?
            raise ValueError(f'Time is over! No enough data were getting! '
                             f'Response length is {self._response_length}, but '
                             f'should be {self._expected_response_length}. '
                             f'Response: {hex_repr(self._response)}')

        # Check command part
        if self._checksum(self._response_cmd[:-1]) != self._response_cmd[-1] \
                or self._response[4] != COMMAND['RESPONSE'] \
                or self._response[:4] != self._request[:4] \
                or self._response_cmd[5:-1] != self._request[5:-1]:
            raise ValueError(f'Response command part failed! '
                             f'Response: {hex_repr(response)}')

        # Check data part
        if self._checksum(self._response_data[:-1]) != self._response_data[-1]:
            raise ValueError(f'Data checksum failed! Data checksum is '
                             f'{hex_repr(self._response_data[-1])}, but should be '
                             f'{hex_repr(self._checksum(self._response_data[:-1]))}. '
                             f'Response: {hex_repr(response)}')

    @property
    def _response_length(self) -> int:
        return self.__response_length

    @property
    def _response_cmd(self) -> bytes:
        return self.__response_cmd

    @property
    def _response_data(self) -> bytes:
        return self.__response_data

    @property
    def _eeprom_data_length(self) -> int:
        return self.__eeprom_data_length

    @property
    def _expected_request_length(self) -> int:
        return self.__expected_request_length

    @property
    def _expected_response_length(self) -> int:
        return self.__expected_response_length

    @property
    def _data_length(self) -> int:
        return self.__data_length

    @_data_length.setter
    def _data_length(self, data_length: int):
        if isinstance(data_length, int) and 0x00 <= data_length <= 0xFF:
            self.__data_length = data_length
        elif not isinstance(data_length, int):
            raise TypeError(f'Data length must be {int}, not {type(data_length)}')
        elif not 0x00 <= data_length <= 0xFF:
            raise ValueError(f'Data length must be in range [0x00, 0xFF], '
                             f'not {data_length}')

    @property
    def _eeprom_addr(self) -> int:
        return self.__eeprom_addr

    @_eeprom_addr.setter
    def _eeprom_addr(self, eeprom_addr: int):
        if isinstance(eeprom_addr, int) and 0x00 <= eeprom_addr <= 0xFF:
            self.__eeprom_addr = eeprom_addr
        elif not isinstance(eeprom_addr, int):
            raise TypeError(f'Node address must be {int}, not {type(eeprom_addr)}')
        elif not 0x00 <= eeprom_addr <= 0xFF:
            raise ValueError(f'Node address must be in range [0x00, 0xFF], '
                             f'not {eeprom_addr}')

    @property
    def _eeprom_data(self) -> bytes:
        return self.__eeprom_data

    @_eeprom_data.setter
    def _eeprom_data(self, eeprom_data: Union[bytes, list, tuple, int]):
        if not isinstance(eeprom_data, (bytes, list, tuple, int)):
            raise TypeError(f'Data must be {bytes}, {list}, {tuple} or {int}, '
                            f'not {type(eeprom_data)}')
        if isinstance(eeprom_data, int):
            eeprom_data = [eeprom_data]
        elif len(eeprom_data) >= EEPROM_SIZE - self._eeprom_addr:    # TODO: EEPROM Addr должен быть задана до!
            raise ValueError(f'Data must not be out of EEPROM.')
        if not all((isinstance(x, int) and 0x00 <= x <= 0xFF) for x in eeprom_data):
            raise TypeError(f'{type(eeprom_data)} items\' data must be {int} '
                            f'and be in range [0x00, 0xFF]')

        self.__eeprom_data = bytes(eeprom_data)
        self.__eeprom_data_length = len(self.__eeprom_data)

    def _pretty_eeprom(self, eeprom: bytes) -> str:
        """
        Compose pretty string from array of hex data (EEPROM) for logging
        """

        pretty_eeprom = ''
        lines, columns = divmod(len(eeprom), 16)
        for line in range(lines):
            pretty_eeprom += '\n\t ' + ' '.join(
                [str(format(byte, '#04x'))
                 for byte in eeprom[line * 16: (line + 1) * 16]]
            )
        if columns:
            pretty_eeprom += '\n\t ' + ' '.join(
                [str(format(byte, '#04x'))
                 for byte in eeprom[lines * 16: len(eeprom)]]
            )

        return pretty_eeprom

    def _checksum(self, data: Union[bytes, list, tuple, set]) -> int:
        """Calculate checksum for data.

        Summarize all bytes (int) in data and return least significant byte.

        Parameters
        ----------
        data : Union[bytes, list, tuple, set]
            Data, for which checksum is calculated. It could be bytes or list,
            tuple, set of int values.

        Returns
        -------
        checksum : int
            Calculated checksum for input data.

        Examples
        --------
        >>> self._checksum(b'\x00\x12\x23\x34\x02\x80\x00')
        235
        >>> self._checksum([0x00, 0x12, 0x23, 0x34, 0x02, 0x80, 0x00])
        235
        """
        if not isinstance(data, (bytes, list, tuple, set)):
            raise TypeError(f'Data must be {bytes}, {list} or {tuple}, '
                            f'not {type(data)}')
        if not all((isinstance(x, int) and 0x00 <= x <= 0xFF) for x in data):
            raise TypeError(f'{type(data)} items\' data must be {int} and be '
                            f'in range [0x00, 0xFF]')

        return sum(data) & 0xFF

    def send_cmd(self,
                 node_addr: Union[bytes, list, tuple],
                 procedure: int) -> bool:
        """Send command to node.

        Parameters
        ----------
        node_addr : Union[bytes, list, tuple]
            Target node address to which send command. It should be 3 bytes or
            3 int (list or tuple).
        procedure : int
            It should be one of the PROCEDURE constant: ``SWITCH_ON``,
            ``STEP_UP``, ``SWITCH_INVERSION``, ``SWITCH_OFF``, ``STEP_DOWN``,
            ``GLOBAL_ON``, ``GLOBAL_OFF``, ``SET_MODE``, ``RESET_MODE`` and etc.

        Returns
        -------
        bool
            True if command was successfully sent, False otherwise.

        Examples
        --------
        >>> self.send_cmd(node_addr=b'\x12\x23\x34', procedure=0x02)
        True
        """

        _LOGGER.debug(f'Calling send_cmd() with Args: {{'
                      f'\'node_addr\': {hex_repr(node_addr)}, '
                      f'\'procedure\': {hex_repr(procedure)}}}')

        try:
            self._node_addr = node_addr
            self._procedure = procedure
        except (TypeError, ValueError) as ex:
            _LOGGER.exception(f'Invalid input parameters: {ex}')
            raise ex

        request = [DEVICE['COMPUTER'],                  # Device ID
                   self._node_addr[0],                   # Address
                   self._node_addr[1],                   # Address
                   self._node_addr[2],                   # Address
                   COMMAND['EXECUTE_PROCEDURE'],        # Command
                   self._procedure,                      # Procedure number
                   0x00]                                # Reserve
        request.append(self._checksum(request[1:]))     # Checksum
        self._request = serial.to_bytes(request)

        _LOGGER.debug(f'Request: {hex_repr(request)}')

        controller_is_busy = True
        time_to_wait = self.__unlock_timeout
        # ==================== RADIO CONTROLLER START ====================
        try:
            self._controller.open()
        except serial.SerialException as ex:
            _LOGGER.error(f'SerialException: Radio controller {self._port_path} '
                          f'is not available! {ex}', exc_info=True)
            raise ex
        else:
            while time_to_wait > 0:
                time_to_wait -= self.__sleep_time
                try:
                    fcntl.flock(self._controller.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    sleep(self.__sleep_time)
                else:
                    _LOGGER.debug(f'Radio controller is available: {self._controller}')

                    # ----- COMMUNICATION START -----
                    # Send request
                    self._controller.reset_output_buffer()
                    self._request_length = self._controller.write(request)
                    # ----- COMMUNICATION END -----

                    if self.__block:
                        sleep(0.3)

                    controller_is_busy = False
                    break
        finally:
            self._controller.close()
        # ==================== RADIO CONTROLLER END ======================

        if controller_is_busy:
            _LOGGER.debug(f'Radio controller {self._controller} is busy')
            raise BlockingIOError(f'Radio controller {self._controller} is busy')

        return True

    def read(self,
             node_addr: Union[bytes, list, tuple],
             eeprom_addr: int,
             data_size: int) -> bytes:
        """
        Read data from Node EEPROM
        Node's examples: relay, dimmer, motor, thermoregulator

        Parameters
        ----------
        node_addr
        eeprom_addr
        data_size

        Returns
        -------

        """

        _LOGGER.debug(f'Calling read() with Args:'
                      f'\'node_addr\': {hex_repr(node_addr)}, '
                      f'\'eeprom_addr\': {hex_repr(eeprom_addr)}, '
                      f'\'data_size\': {hex_repr(data_size)}')

        try:
            self._node_addr = node_addr
            self._data_length = data_size
            self._eeprom_addr = eeprom_addr
        except (TypeError, ValueError) as ex:
            _LOGGER.exception(f'Invalid input parameters: {ex}')
            raise ex

        request = [DEVICE['COMPUTER'],                  # Device ID
                  self._node_addr[0],                    # Address
                  self._node_addr[1],                    # Address
                  self._node_addr[2],                    # Address
                  COMMAND['READ_EEPROM'],               # Command
                  self._data_length,                     # Size of data to read
                  self._eeprom_addr]                     # Start address EEPROM, where read
        request.append(self._checksum(request[1:]))     # Checksum
        self._request = serial.to_bytes(request)

        _LOGGER.debug(f'Request: {hex_repr(request)}')

        controller_is_busy = False
        time_to_wait = self.__unlock_timeout

        # ==================== RADIO CONTROLLER START ====================
        try:
            self._controller.open()
        except serial.SerialException as ex:
            _LOGGER.error(f'SerialException: Radio controller {self._port_path} '
                          f'is not available! {ex}', exc_info=True)
            raise ex
        else:
            while time_to_wait > 0:
                time_to_wait -= self.__sleep_time
                try:
                    fcntl.flock(self._controller.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    sleep(self.__sleep_time)
                else:
                    _LOGGER.debug(f'Radio controller is available: {self._controller}')

                    # ----- COMMUNICATION START -----
                    # Send request
                    self._controller.reset_output_buffer()

                    try:
                        self._request_length = self._controller.write(request)
                    except ValueError as ex:
                        _LOGGER.exception(f'Error while sending request: {ex}')
                        raise ex

                    # Receive response
                    time1 = time()
                    self._controller.reset_input_buffer()
                    try:
                        self._response = self._controller.read(self._expected_response_length)
                    except ValueError as ex:
                        _LOGGER.exception(f'Error while receiving response: {ex}')
                        raise ex
                    time2 = time()
                    # ----- COMMUNICATION END -----

                    if self.__block:
                        sleep(0.1)

                    controller_is_busy = True
                    break
        finally:
            self._controller.close()
        # ==================== RADIO CONTROLLER END ======================

        if not controller_is_busy:
            _LOGGER.info(f'Radio controller {self._controller} is busy')
            raise BlockingIOError(f'Radio controller {self._controller} is busy')
        _LOGGER.debug(f'Response transfer time: {time2 - time1} sec')
        _LOGGER.debug(f'Request length -> {self._request_length}. '
                      f'Response length -> {self._response_length}.')
        _LOGGER.debug(f'EEPROM data: {hex_repr(self._response_data[:-1])}')
        # _LOGGER.debug(f'EEPROM data: {self._pretty_eeprom(self._response_data[:-1])}')

        return self._response_data[:-1]     # Last byte is check sum

    def write(self,
              node_addr: Union[bytes, list, tuple],
              eeprom_addr: int,
              data: Union[int, bytes, list, tuple]) -> bool:
        """
        Write data to Node EEPROM.
        Node's examples: relay, dimmer, motor, thermoregulator

        Parameters
        ----------
        node_addr
        eeprom_addr
        data

        Returns
        -------

        """

        _LOGGER.debug(f'Calling write() with Args:'
                      f'\'node_addr\': {hex_repr(node_addr)}, '
                      f'\'eeprom_addr\': {hex_repr(eeprom_addr)}, '
                      f'\'data\': {hex_repr(data)}')

        try:
            self._node_addr = node_addr
            self._eeprom_addr = eeprom_addr
            self._eeprom_data = data
        except (TypeError, ValueError) as ex:
            _LOGGER.exception(f'Invalid input parameters: {ex}')
            raise ex

        request = [DEVICE['COMPUTER'],                  # Device ID
                   node_addr[0],                        # Address
                   node_addr[1],                        # Address
                   node_addr[2],                        # Address
                   COMMAND['WRITE_EEPROM'],             # Command
                   self._eeprom_data_length,            # Size of data to write
                   eeprom_addr]                         # Start EEPROM address, where write
        request.append(self._checksum(request[1:]))     # Checksum of command
        request += self._eeprom_data
        request.append(self._checksum(self._eeprom_data))    # Checksum of data
        self._request = serial.to_bytes(request)

        _LOGGER.debug(f'Request: {hex_repr(request)}')

        controller_is_busy = True
        time_to_wait = self.__unlock_timeout
        # ==================== RADIO CONTROLLER START ====================
        try:
            self._controller.open()
        except serial.SerialException as ex:
            _LOGGER.error(f'SerialException: Radio controller {self._port_path} '
                          f'is not available! {ex}', exc_info=True)
            raise ex
        else:
            while time_to_wait > 0:
                time_to_wait -= self.__sleep_time
                try:
                    fcntl.flock(self._controller.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                except BlockingIOError:
                    sleep(self.__sleep_time)
                else:
                    _LOGGER.debug(f'Radio controller is available: {self._controller}')

                    # ----- COMMUNICATION START -----
                    # Send request
                    self._controller.reset_output_buffer()
                    self._request_length = self._controller.write(request)
                    # ----- COMMUNICATION END -----

                    if self.__block:
                        expected_block_time = 0.3 + self._request_length / 280
                        sleep(expected_block_time)

                    controller_is_busy = False
                    break
        finally:
            self._controller.close()
        # ==================== RADIO CONTROLLER END ======================

        if controller_is_busy:
            _LOGGER.debug(f'Radio controller {self._controller} is busy')
            raise BlockingIOError(f'Radio controller {self._controller} is busy')

        return True



logging_config = dict(version=1,
                      disable_existing_loggers=False,
                      formatters={'verbose': {'format': '[%(asctime)s] - '
                                                        '%(levelname)-8s - '
                                                        '%(filename)s:%(lineno)d '
                                                        '(%(funcName)s) - '
                                                        '%(message)s'}},
                      handlers={'console': {'class': 'logging.StreamHandler',
                                            'level': 'DEBUG',
                                            'formatter': 'verbose',
                                            'stream': 'ext://sys.stdout'},
                                # 'info_file_handler': {'class': 'logging.handlers.RotatingFileHandler',
                                #                       'level': 'INFO',
                                #                       'formatter': 'verbose',
                                #                       'filename': 'info.log',
                                #                       'maxBytes': 10485760,     # 10MB
                                #                       'backupCount': 20,
                                #                       'encoding': 'utf8'},
                                # 'error_file_handler': {'class': 'logging.handlers.RotatingFileHandler',
                                #                        'level': 'ERROR',
                                #                        'formatter': 'verbose',
                                #                        'filename': 'errors.log',
                                #                        'maxBytes': 10485760,    # 10MB
                                #                        'backupCount': 20,
                                #                        'encoding': 'utf8'}
                                },
                      loggers={'controller_module': {'level': 'ERROR',
                                                     'handlers': ['console'],
                                                     'propagate': False},
                               'relay_module': {'level': 'ERROR',
                                                'handlers': ['console'],
                                                'propagate': False}},
                      root={'level': 'ERROR',
                            'handlers': ['console']}
                      )


# General eeprom addresses for settings
EEPROM_HARDWARE_TYPE = 0xE8
EEPROM_FIRMWARE_VERSION = 0xF0
EEPROM_SIZE = 256


# Devices types
DEVICE = dict(COMPUTER=0x00,
              GIDROLOCK=0x01,
              DELUMO=0x02,
              THERMOREGULATOR=0x03)

# Commands
COMMAND = dict(WRITE_EEPROM=0x01,
               READ_EEPROM=0x02,
               RESPONSE=0x03,
               EXECUTE_PROCEDURE=0x07)

HARDWARE_TYPE = dict(RELAY=0x5E,
                     DIMMER=0x5D,
                     MOTOR=0x5C)


from setuptools import setup


def readme():
    with open('README.rst', 'r') as f:
        return f.read()


setup(
    name='DelumoWave',
    version='0.1.1',
    description='API for Delumo radio controller',
    long_description=readme(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Topic :: Home Automation',
        'Topic :: System :: Hardware',
        'Topic :: System :: Hardware :: Hardware Drivers',
    ],
    license='MIT',
    author='Roman Alexeev',
    author_email='roman@alexeyev.su',
    url='https://bitbucket.org/delumo/DelumoWave',
    packages=['delumowave', 'delumowave.config'],
    install_requires=[
        'pyserial >= 3.4',
    ],
    include_package_data=True,
    zip_safe=False,
)



#!/usr/bin/env python3

import logging
from typing import Union

from delumowave.controller import DelumoWaveController, hex_repr
from delumowave.config.relay import (RELAY, BRIGHTNESS, STATE, DIMMING_STEP)

_LOGGER = logging.getLogger(__name__)


class DelumoWaveRelay(DelumoWaveController):
    def __init__(self,
                 node_addr: Union[bytes, list, tuple],
                 node_version: int = max(RELAY, key=int),
                 port_path: str = '/dev/ttyUSB0',
                 unlock_timeout: float = 5.0,
                 sleep_time: float = 0.1,
                 block: bool = False,
                 logging_level: Union[str, int, None] = None):
        """
        This class provide methods to manage relay/dimmer.

        Parameters
        ----------
        node_addr : Union[bytes, list, tuple]
            Node address. Three latest meaningful bytes. For example:
            ``b'\\x12\\xA7\\xE4'``, ``[173, 34, 211]``, ``[0x7A, 0x19, 0xF0]``.
        node_version : int
            Node firmware version. One byte from 0x00 to 0xFF.
            By default, latest available firmware version.
        port_path : str, optional
            Device name: depending on operating system. e.g. ``/dev/ttyUSB0``,
            ``/dev/ttyAMA0`` on GNU/Linux or ``COM3`` on Windows.
            By default, ``/dev/ttyAMA0`` is used.
        unlock_timeout : float
        sleep_time : float
        block : float
        logging_level : Union[str, int, None], optional
            Available logging levels: ``DEBUG``, ``INFO``, ``WARNING``,
            ``ERROR``, ``CRITICAL``, ``NOTSET``. Also it could be written by
            numeric values.
            By default, logging is disabled.
        """

        if logging_level is None:
            _LOGGER.disabled = True
        elif isinstance(logging_level, str):
            if logging_level.upper() in ('DEBUG', 'INFO', 'WARNING', 'ERROR',
                                         'CRITICAL', 'NOTSET'):
                _LOGGER.setLevel(logging_level.upper())
        elif isinstance(logging_level, int):
            _LOGGER.setLevel(logging_level)
        else:
            _LOGGER.setLevel('ERROR')

        _LOGGER.debug(f'DelumoWaveRelay __init__() with Args: {{'
                      f'\'node_addr\': {hex_repr(node_addr)}, '
                      f'\'port_path\': {port_path}}}')

        super().__init__(port_path=port_path,
                         unlock_timeout=unlock_timeout,
                         sleep_time=sleep_time,
                         block=block,
                         logging_level=logging_level)

        self._node_version = node_version
        self._node_addr = node_addr

        #: Relay/dimmer constants for appropriate firmware version
        self.RELAY = RELAY[self._node_version]

    def invert(self) -> bool:
        """
        Invert state (``ON`` - 0x55 or ``OFF`` - 0xFF) of relay/dimmer

        Returns
        -------
        bool
            True if the command to invert state of relay/dimmer was
            successfully sent, False otherwise.
        """

        _LOGGER.debug('Calling invert()')

        try:
            return self.send_cmd(node_addr=self._node_addr,
                                 procedure=self.RELAY['PROCEDURE']['SWITCH_INVERSION'])
        except KeyError as ex:
            _LOGGER.error(f'{ex} is not supported for '
                          f'{hex_repr(self._node_version)} relay/dimmer '
                          f'firmware version. ', exc_info=True)
            return False
        # except Exception as ex:
        #     _LOGGER.exception(ex, exc_info=True)

    def relay_on(self) -> bool:
        """
        Switch on relay/dimmer by executing command with procedure ``SWITCH_ON``

        Returns
        -------
        bool
            True if the command to switch on relay/dimmer was successfully sent,
            False otherwise.
        """

        _LOGGER.debug('Calling relay_on()')

        try:
            return self.send_cmd(node_addr=self._node_addr,
                                 procedure=self.RELAY['PROCEDURE']['SWITCH_ON'])
        except KeyError as ex:
            _LOGGER.error(f'{ex} is not supported for '
                          f'{hex_repr(self._node_version)} relay/dimmer '
                          f'firmware version. ', exc_info=True)
            return False

    def relay_off(self) -> bool:
        """
        Switch off relay/dimmer by executing command with procedure
        ``SWITCH_OFF``

        Returns
        -------
        bool
            True if the command to switch off relay/dimmer was successfully
            sent, False otherwise.
        """

        _LOGGER.debug('Calling relay_off()')

        try:
            return self.send_cmd(node_addr=self._node_addr,
                                 procedure=self.RELAY['PROCEDURE']['SWITCH_OFF'])
        except KeyError as ex:
            _LOGGER.error(f'{ex} is not supported for '
                          f'{hex_repr(self._node_version)} relay/dimmer '
                          f'firmware version. ', exc_info=True)
            return False

    def dimmer_on(self) -> bool:
        """
        Switch on dimmer by writing state to EEPROM
        """

        _LOGGER.debug('Dimmer on')

        return self.send_cmd(node_addr=self._node_addr,
                             procedure=self.RELAY['PROCEDURE']['GLOBAL_ON'])

    def dimmer_off(self) -> bool:
        """
        Switch off dimmer by writing state to EEPROM
        """

        _LOGGER.debug('Dimmer off')

        return self.send_cmd(node_addr=self._node_addr,
                             procedure=self.RELAY['PROCEDURE']['GLOBAL_OFF'])

    @property
    def time_to_relay_off(self) -> int:
        """
        Get time to relay/dimmer off in hours
        """

        _LOGGER.debug('Get time to relay off')
        time_to_relay_off = self.read(node_addr=self._node_addr,
                                      eeprom_addr=self.RELAY['EEPROM']['TIME_RELAY_OFF'],
                                      data_size=0x01)[0]
        _LOGGER.debug(f'Time to relay off -> {hex_repr(time_to_relay_off)}')

        return time_to_relay_off

    @time_to_relay_off.setter
    def time_to_relay_off(self, time_in_hours: int):
        """
        Set time to relay/dimmer off in hours
        """

        _LOGGER.debug(f'Set time to relay off {time_in_hours}')
        status = self.write(node_addr=self._node_addr,
                            eeprom_addr=self.RELAY['EEPROM']['TIME_RELAY_OFF'],
                            data=time_in_hours)
        _LOGGER.debug(f'Time to relay off -> {hex_repr(time_in_hours)}. '
                      f'{"OK!" if status else "Failed!"}')

    @property
    def hardware_type(self) -> int:
        """
        Get hardware type of node (relay or dimmer)
        """

        _LOGGER.debug('Get hardware type')
        hardware_type = self.read(node_addr=self._node_addr,
                                  eeprom_addr=self.RELAY['EEPROM']['HARDWARE_TYPE'],
                                  data_size=0x01)[0]
        _LOGGER.debug(f'Relay hardware type -> {hex_repr(hardware_type)}')

        return hardware_type

    @hardware_type.setter
    def hardware_type(self, hardware_type):
        """
        Set hardware type of node (relay or dimmer)
        """

        _LOGGER.debug(f'Set hardware type {hardware_type}')
        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['HARDWARE_TYPE'],
                   data=hardware_type)
        _LOGGER.debug(f'Relay hardware type -> {hex_repr(hardware_type)}')

    @property
    def firmware_version(self) -> int:
        """
        Get firmware version of node
        """

        _LOGGER.debug('Get firmware version')
        firmware_version = self.read(node_addr=self._node_addr,
                                     eeprom_addr=self.RELAY['EEPROM']['FIRMWARE_VERSION'],
                                     data_size=0x01)[0]
        _LOGGER.debug(f'Relay firmware version -> {hex_repr(firmware_version)}')

        return firmware_version

    @firmware_version.setter
    def firmware_version(self, firmware_version):
        """
        Set firmware version of node
        """

        _LOGGER.debug('Get firmware version')
        self.write(
            node_addr=self._node_addr,
            eeprom_addr=self.RELAY['EEPROM']['FIRMWARE_VERSION'],
            data=firmware_version)
        _LOGGER.debug(f'Relay firmware version -> {hex_repr(firmware_version)}')

    @property
    def state(self) -> bool:
        """
        Get state of node - on (0x55) or off (0xFF)
        """

        _LOGGER.debug('Get state')
        state_byte = self.read(node_addr=self._node_addr,
                               eeprom_addr=self.RELAY['EEPROM']['STATE'],
                               data_size=0x01)[0]
        _LOGGER.debug(f'Relay state -> {hex_repr(state_byte)}')
        state = next((x for x, y in STATE.items() if y == state_byte), None)
        _LOGGER.debug(f'Relay state -> {state}')

        return state

    @state.setter
    def state(self, state: bool):
        """
        Set state of node - ON (0x55) or OFF (0xFF)
        """

        _LOGGER.debug(f'Set state {state}')
        state_byte = STATE[state]
        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['STATE'],
                   data=state_byte)
        _LOGGER.debug(f'Relay state -> {state} ({state_byte})')

    @property
    def brightness_level(self) -> int:
        """
        Get brightness level of dimmer
        """

        _LOGGER.debug('Get brightness_level')
        brightness_level = self.read(node_addr=self._node_addr,
                                     eeprom_addr=self.RELAY['EEPROM']['BRIGHTNESS_LEVEL'],
                                     data_size=0x01)[0]
        _LOGGER.debug(f'Relay brightness level -> {hex_repr(brightness_level)}')

        return brightness_level

    @brightness_level.setter
    def brightness_level(self, brightness_level: int):
        """
        Set brightness level of dimmer
        """

        _LOGGER.debug(f'Set brightness_level {brightness_level}')
        if brightness_level > BRIGHTNESS['MAX']:
            brightness_level = BRIGHTNESS['MAX']
        if brightness_level < BRIGHTNESS['MIN']:
            brightness_level = BRIGHTNESS['MIN']

        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['BRIGHTNESS_LEVEL'],
                   data=brightness_level)
        _LOGGER.debug(f'Relay brightness level -> {hex_repr(brightness_level)}')

    @property
    def brightness_level_comfort(self) -> int:
        """
        Get brightness level comfort of dimmer
        """

        _LOGGER.debug('Get brightness level comfort')
        brightness_level_comfort = self.read(node_addr=self._node_addr,
                                             eeprom_addr=self.RELAY['EEPROM']['BRIGHTNESS_LEVEL_COMFORT'],
                                             data_size=0x01)[0]
        _LOGGER.debug(f'Relay brightness level comfort -> {hex_repr(brightness_level_comfort)}')

        return brightness_level_comfort

    @brightness_level_comfort.setter
    def brightness_level_comfort(self, brightness_level_comfort: int):
        """
        Set brightness level comfort of dimmer
        """

        _LOGGER.debug(f'Set brightness level comfort {brightness_level_comfort}')
        if brightness_level_comfort > BRIGHTNESS['MAX']:
            brightness_level_comfort = BRIGHTNESS['MAX']
        if brightness_level_comfort < BRIGHTNESS['MIN']:
            brightness_level_comfort = BRIGHTNESS['MIN']

        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['BRIGHTNESS_LEVEL_COMFORT'],
                   data=brightness_level_comfort)
        _LOGGER.debug(f'Relay brightness level comfort -> {hex_repr(brightness_level_comfort)}')

    @property
    def dimming_speed(self) -> int:
        """
        Get dimming speed
        """

        _LOGGER.debug('Get dimming speed')
        dimming_speed = self.read(node_addr=self._node_addr,
                                  eeprom_addr=self.RELAY['EEPROM']['DIMMING_SPEED'],
                                  data_size=0x01)[0]
        _LOGGER.debug(f'Relay dimming speed -> {hex_repr(dimming_speed)}')

        return dimming_speed

    @dimming_speed.setter
    def dimming_speed(self, dimming_speed: int):
        """
        Set dimming speed
        """

        _LOGGER.debug(f'Set dimming speed {dimming_speed}')
        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['DIMMING_SPEED'],
                   data=dimming_speed)
        _LOGGER.debug(f'Relay dimming speed -> {hex_repr(dimming_speed)}')

    @property
    def dimming_step(self) -> int:
        """
        Get dimming step
        """

        _LOGGER.debug('Get dimming step')
        dimming_step = self.read(node_addr=self._node_addr,
                                 eeprom_addr=self.RELAY['EEPROM']['DIMMING_STEP'],
                                 data_size=0x01)[0]
        _LOGGER.debug(f'Relay dimming step -> {hex_repr(dimming_step)}')

        return dimming_step

    @dimming_step.setter
    def dimming_step(self, dimming_step: int):
        """
        Set dimming step
        """

        _LOGGER.debug(f'Set dimming step {dimming_step}')
        if dimming_step > DIMMING_STEP['MAX']:
            dimming_step = DIMMING_STEP['MAX']
        if dimming_step < DIMMING_STEP['MIN']:
            dimming_step = DIMMING_STEP['MIN']

        self.write(node_addr=self._node_addr,
                   eeprom_addr=self.RELAY['EEPROM']['DIMMING_STEP'],
                   data=dimming_step)
        _LOGGER.debug(f'Relay dimming step -> {hex_repr(dimming_step)}')
